/*******************************************************************************
  MPLAB Harmony Application Source File
  
  Company:
    Microchip Technology Inc.
  
  File Name:
    app.c

  Summary:
    This file contains the source code for the MPLAB Harmony application.

  Description:
    This file contains the source code for the MPLAB Harmony application.  It 
    implements the logic of the application's state machine and it may call 
    API routines of other MPLAB Harmony modules in the system, such as drivers,
    system services, and middleware.  However, it does not call any of the
    system interfaces (such as the "Initialize" and "Tasks" functions) of any of
    the modules in the system or make any assumptions about when those functions
    are called.  That is the responsibility of the configuration-specific system
    files.
 *******************************************************************************/

// *****************************************************************************
// *****************************************************************************
// Section: Included Files 
// *****************************************************************************
// *****************************************************************************

#include "app.h"
#include <string.h>
#include <math.h>
// *****************************************************************************
// *****************************************************************************
// Section: Global Data Definitions
// *****************************************************************************
// *****************************************************************************

#define ACCL_ADDR  0x19
#define GYRO_ADDR  0x69
#define MAG_ADDR   0x13

uint8_t CACHE_ALIGN switchPromptUSB[] = "\r\nPUSH BUTTON PRESSED";

uint8_t CACHE_ALIGN cdcReadBuffer[APP_READ_BUFFER_SIZE];
uint8_t CACHE_ALIGN cdcWriteBuffer[APP_READ_BUFFER_SIZE];

static uint8_t TxData[2] = {0};
static uint8_t RxData[6] = {0};

float xAccl = 0.0;
float yAccl = 0.0;
float zAccl = 0.0;
float xGyro = 0.0;
float yGyro = 0.0;
float zGyro = 0.0;
float xMag  = 0.0;
float yMag  = 0.0;
float zMag  = 0.0;

void DelayMs(uint16_t);
// *****************************************************************************
/* Application Data

  Summary:
    Holds application data

  Description:
    This structure holds the application's data.

  Remarks:
    This structure should be initialized by the APP_Initialize function.
    
    Application strings and buffers are be defined outside this structure.
*/

APP_DATA appData;

// *****************************************************************************
// *****************************************************************************
// Section: Application Callback Functions
// *****************************************************************************
// *****************************************************************************

void APP_I2CEventHandler (
    DRV_I2C_TRANSFER_EVENT event,
    DRV_I2C_TRANSFER_HANDLE transferHandle,
    uintptr_t context
)
{
    APP_TRANSFER_STATUS* transferStatus = (APP_TRANSFER_STATUS*)context;

    if (event == DRV_I2C_TRANSFER_EVENT_COMPLETE)
    {
        if (transferStatus)
        {
            *transferStatus = APP_TRANSFER_STATUS_SUCCESS;
        }
    }
    else
    {
        if (transferStatus)
        {
            *transferStatus = APP_TRANSFER_STATUS_ERROR;
        }
    }
}

/*******************************************************
 * USB CDC Device Events - Application Event Handler
 *******************************************************/

USB_DEVICE_CDC_EVENT_RESPONSE APP_USBDeviceCDCEventHandler
(
    USB_DEVICE_CDC_INDEX index,
    USB_DEVICE_CDC_EVENT event,
    void * pData,
    uintptr_t userData
)
{
    APP_DATA * appDataObject;
    USB_CDC_CONTROL_LINE_STATE * controlLineStateData;
    USB_DEVICE_CDC_EVENT_DATA_READ_COMPLETE * eventDataRead;
    
    appDataObject = (APP_DATA *)userData;

    switch(event)
    {
        case USB_DEVICE_CDC_EVENT_GET_LINE_CODING:

            /* This means the host wants to know the current line
             * coding. This is a control transfer request. Use the
             * USB_DEVICE_ControlSend() function to send the data to
             * host.  */

            USB_DEVICE_ControlSend(appDataObject->deviceHandle,
                    &appDataObject->getLineCodingData, sizeof(USB_CDC_LINE_CODING));

            break;

        case USB_DEVICE_CDC_EVENT_SET_LINE_CODING:

            /* This means the host wants to set the line coding.
             * This is a control transfer request. Use the
             * USB_DEVICE_ControlReceive() function to receive the
             * data from the host */

            USB_DEVICE_ControlReceive(appDataObject->deviceHandle,
                    &appDataObject->setLineCodingData, sizeof(USB_CDC_LINE_CODING));

            break;

        case USB_DEVICE_CDC_EVENT_SET_CONTROL_LINE_STATE:

            /* This means the host is setting the control line state.
             * Read the control line state. We will accept this request
             * for now. */

            controlLineStateData = (USB_CDC_CONTROL_LINE_STATE *)pData;
            appDataObject->controlLineStateData.dtr = controlLineStateData->dtr;
            appDataObject->controlLineStateData.carrier = controlLineStateData->carrier;

            USB_DEVICE_ControlStatus(appDataObject->deviceHandle, USB_DEVICE_CONTROL_STATUS_OK);

            break;

        case USB_DEVICE_CDC_EVENT_SEND_BREAK:

            /* This means that the host is requesting that a break of the
             * specified duration be sent. Read the break duration */

            appDataObject->breakData = ((USB_DEVICE_CDC_EVENT_DATA_SEND_BREAK *)pData)->breakDuration;
            
            /* Complete the control transfer by sending a ZLP  */
            USB_DEVICE_ControlStatus(appDataObject->deviceHandle, USB_DEVICE_CONTROL_STATUS_OK);
            
            break;

        case USB_DEVICE_CDC_EVENT_READ_COMPLETE:

            /* This means that the host has sent some data*/
            eventDataRead = (USB_DEVICE_CDC_EVENT_DATA_READ_COMPLETE *)pData;
            
            if(eventDataRead->status != USB_DEVICE_CDC_RESULT_ERROR)
            {
                appDataObject->isReadComplete = true;
                
                appDataObject->numBytesRead = eventDataRead->length; 
            }
            break;

        case USB_DEVICE_CDC_EVENT_CONTROL_TRANSFER_DATA_RECEIVED:

            /* The data stage of the last control transfer is
             * complete. For now we accept all the data */

            USB_DEVICE_ControlStatus(appDataObject->deviceHandle, USB_DEVICE_CONTROL_STATUS_OK);
            break;

        case USB_DEVICE_CDC_EVENT_CONTROL_TRANSFER_DATA_SENT:

            /* This means the GET LINE CODING function data is valid. We don't
             * do much with this data in this demo. */
            break;

        case USB_DEVICE_CDC_EVENT_WRITE_COMPLETE:

            /* This means that the data write got completed. We can schedule
             * the next read. */

            appDataObject->isWriteComplete = true;
            break;

        default:
            break;
    }

    return USB_DEVICE_CDC_EVENT_RESPONSE_NONE;
}

/***********************************************
 * Application USB Device Layer Event Handler.
 ***********************************************/
void APP_USBDeviceEventHandler 
(
    USB_DEVICE_EVENT event, 
    void * eventData, 
    uintptr_t context 
)
{
    USB_DEVICE_EVENT_DATA_CONFIGURED *configuredEventData;

    switch(event)
    {
        case USB_DEVICE_EVENT_SOF:

            /* This event is used for switch debounce. This flag is reset
             * by the switch process routine. */
            appData.sofEventHasOccurred = true;
            
            break;

        case USB_DEVICE_EVENT_RESET:

            appData.isConfigured = false;

            break;

        case USB_DEVICE_EVENT_CONFIGURED:

            /* Check the configuration. We only support configuration 1 */
            configuredEventData = (USB_DEVICE_EVENT_DATA_CONFIGURED*)eventData;
            
            if ( configuredEventData->configurationValue == 1)
            {
         
                /* Register the CDC Device application event handler here.
                 * Note how the appData object pointer is passed as the
                 * user data */

                USB_DEVICE_CDC_EventHandlerSet(USB_DEVICE_CDC_INDEX_0, APP_USBDeviceCDCEventHandler, (uintptr_t)&appData);

                /* Mark that the device is now configured */
                appData.isConfigured = true;
            }
            
            break;

        case USB_DEVICE_EVENT_POWER_DETECTED:

            /* VBUS was detected. We can attach the device */
            USB_DEVICE_Attach(appData.deviceHandle);
            
            break;

        case USB_DEVICE_EVENT_POWER_REMOVED:
            
            /* VBUS is not available. We can detach the device */
            USB_DEVICE_Detach(appData.deviceHandle);
            
            appData.isConfigured = false;
                    
            break;

        case USB_DEVICE_EVENT_SUSPENDED:
          
            break;

        case USB_DEVICE_EVENT_RESUMED:
        case USB_DEVICE_EVENT_ERROR:
        default:
            
            break;
    }
}

// *****************************************************************************
// *****************************************************************************
// Section: Application Local Functions
// *****************************************************************************
// *****************************************************************************



/*****************************************************
 * This function is called in every step of the
 * application state machine.
 *****************************************************/

bool APP_StateReset(void)
{
    /* This function returns true if the device
     * was reset  */

    bool retVal;

    if(appData.isConfigured == false)
    {
        appData.state = APP_STATE_WAIT_FOR_CONFIGURATION;
        appData.readTransferHandle = USB_DEVICE_CDC_TRANSFER_HANDLE_INVALID;
        appData.writeTransferHandle = USB_DEVICE_CDC_TRANSFER_HANDLE_INVALID;
        appData.isReadComplete = true;
        appData.isWriteComplete = true;
        retVal = true;
    }
    else
    {
        retVal = false;
    }

    return(retVal);
}

// *****************************************************************************
// *****************************************************************************
// Section: Application Initialization and State Machine Functions
// *****************************************************************************
// *****************************************************************************

/*******************************************************************************
  Function:
    void APP_Initialize(void)

  Remarks:
    See prototype in app.h.
 */

void APP_Initialize(void)
{
    /* Place the App state machine in its initial state. */
    appData.state = APP_STATE_INIT;
    
    /* Device Layer Handle  */
    appData.deviceHandle = USB_DEVICE_HANDLE_INVALID ;

    /* Device configured status */
    appData.isConfigured = false;

    /* Initial get line coding state */
    appData.getLineCodingData.dwDTERate = 9600;
    appData.getLineCodingData.bParityType = 0;
    appData.getLineCodingData.bParityType = 0;
    appData.getLineCodingData.bDataBits = 8;

    /* Read Transfer Handle */
    appData.readTransferHandle = USB_DEVICE_CDC_TRANSFER_HANDLE_INVALID;

    /* Write Transfer Handle */
    appData.writeTransferHandle = USB_DEVICE_CDC_TRANSFER_HANDLE_INVALID;

    /* Initialize the read complete flag */
    appData.isReadComplete = true;

    /*Initialize the write complete flag*/
    appData.isWriteComplete = true;

    /* Reset other flags */
    appData.sofEventHasOccurred = false;

    /* Set up the read buffer */
    appData.cdcReadBuffer = &cdcReadBuffer[0];

    /* Set up the read buffer */
    appData.cdcWriteBuffer = &cdcWriteBuffer[0];       
}


/******************************************************************************
  Function:
    void APP_Tasks(void)

  Remarks:
    See prototype in app.h.
 */

void APP_Tasks(void)
{
    /* Update the application state machine based
     * on the current state */
//    int i;
    
    switch(appData.state)
    {
        case APP_STATE_INIT:

            /* Open the device layer */
            appData.deviceHandle = USB_DEVICE_Open( USB_DEVICE_INDEX_0, DRV_IO_INTENT_READWRITE );

            if(appData.deviceHandle != USB_DEVICE_HANDLE_INVALID)
            {
                /* Register a callback with device layer to get event notification (for end point 0) */
                USB_DEVICE_EventHandlerSet(appData.deviceHandle, APP_USBDeviceEventHandler, 0);

                appData.state = APP_STATE_WAIT_FOR_CONFIGURATION;
            }
            else
            {
                /* The Device Layer is not ready to be opened. We should try
                 * again later. */
            }

            break;

        case APP_STATE_WAIT_FOR_CONFIGURATION:

            /* Check if the device was configured */
            if(appData.isConfigured)
            {
                /* If the device is configured then lets start reading */
                appData.state = APP_STATE_I2C_INIT;
            }
            
            break;
            
///////////////////////////////////////////////////////    

        case APP_STATE_I2C_INIT:
          
            /* Open the I2C Driver */
            appData.drvI2CHandle = DRV_I2C_Open( DRV_I2C_INDEX_0, DRV_IO_INTENT_READWRITE );
         
            /* Register the I2C Driver event Handler */
            DRV_I2C_TransferEventHandlerSet(
                appData.drvI2CHandle,
                APP_I2CEventHandler,
                (uintptr_t)&appData.transferStatus
            );
                          
            appData.state = APP_STATE_BMX055_SET_ACCL_RANGE;
             
            break;
            
        case APP_STATE_BMX055_SET_ACCL_RANGE:
            
            appData.transferStatus = APP_TRANSFER_STATUS_IN_PROGRESS; 
            
            TxData[0] =0x0f;  // Select PMU_Range register
            TxData[1]= 0x03;  // Range = +/- 2g
            
            DRV_I2C_WriteTransferAdd(appData.drvI2CHandle, ACCL_ADDR,
                (void *)&TxData, 2,
                &appData.transferHandle
            );

            appData.state = APP_STATE_BMX055_SET_ACCL_BW;
            
            break;

        case APP_STATE_BMX055_SET_ACCL_BW:
 
            if (appData.transferStatus == APP_TRANSFER_STATUS_SUCCESS)
            {
                appData.transferStatus = APP_TRANSFER_STATUS_IN_PROGRESS; 
            
                TxData[0]= 0x10;  // Select PMU_BW register
                TxData[1]= 0x08;  // Bandwidth = 7.81 Hz
            
                DRV_I2C_WriteTransferAdd(appData.drvI2CHandle, ACCL_ADDR,
                    (void *)&TxData, 2,
                    &appData.transferHandle
                );
                appData.state = APP_STATE_BMX055_SET_ACCL_LPW;
            }
             
            break;
            
        case APP_STATE_BMX055_SET_ACCL_LPW:
 
            if (appData.transferStatus == APP_TRANSFER_STATUS_SUCCESS)
            {
                appData.transferStatus = APP_TRANSFER_STATUS_IN_PROGRESS; 
            
                TxData[0]= 0x11;  // Select PMU_LPW register
                TxData[1]= 0x00;  // Normal mode, Sleep duration = 0.5ms
            
                DRV_I2C_WriteTransferAdd(appData.drvI2CHandle, ACCL_ADDR,
                    (void *)&TxData, 2,
                    &appData.transferHandle
                );
                appData.state = APP_STATE_BMX055_SET_GYRO_RANGE;
            }
             
            break;

        case APP_STATE_BMX055_SET_GYRO_RANGE:
            
            if (appData.transferStatus == APP_TRANSFER_STATUS_SUCCESS)
            {            
                appData.transferStatus = APP_TRANSFER_STATUS_IN_PROGRESS; 
            
                TxData[0] =0x0f;  // Select Range register
                TxData[1]= 0x04;  // Full scale = +/- 125 degree/s
            
                DRV_I2C_WriteTransferAdd(appData.drvI2CHandle, GYRO_ADDR,
                    (void *)&TxData, 2,
                    &appData.transferHandle
                );

                appData.state = APP_STATE_BMX055_SET_GYRO_BW;
            }
            
            break;

        case APP_STATE_BMX055_SET_GYRO_BW:
 
            if (appData.transferStatus == APP_TRANSFER_STATUS_SUCCESS)
            {
                appData.transferStatus = APP_TRANSFER_STATUS_IN_PROGRESS; 
            
                TxData[0]= 0x10;  // Select Bandwidth register
                TxData[1]= 0x07;  // ODR = 100 Hz
            
                DRV_I2C_WriteTransferAdd(appData.drvI2CHandle, GYRO_ADDR,
                    (void *)&TxData, 2,
                    &appData.transferHandle
                );
                appData.state = APP_STATE_BMX055_SET_GYRO_LPM1;
            }
             
            break;
            
        case APP_STATE_BMX055_SET_GYRO_LPM1:
 
            if (appData.transferStatus == APP_TRANSFER_STATUS_SUCCESS)
            {
                appData.transferStatus = APP_TRANSFER_STATUS_IN_PROGRESS; 
            
                TxData[0]= 0x11;  // Select LPM1 register
                TxData[1]= 0x00;  // Normal mode, Sleep duration = 2ms
            
                DRV_I2C_WriteTransferAdd(appData.drvI2CHandle, GYRO_ADDR,
                    (void *)&TxData, 2,
                 &appData.transferHandle
                );
                appData.state = APP_STATE_BMX055_SET_MAG_RESET1;
                
            }
             
            break;
            
            case APP_STATE_BMX055_SET_MAG_RESET1:
 
            if (appData.transferStatus == APP_TRANSFER_STATUS_SUCCESS)
            {
                appData.transferStatus = APP_TRANSFER_STATUS_IN_PROGRESS; 
            
                TxData[0]= 0x4b;  // Select Mag register
                TxData[1]= 0x83;  // Soft reset
            
                DRV_I2C_WriteTransferAdd(appData.drvI2CHandle, MAG_ADDR,
                    (void *)&TxData, 2,
                    &appData.transferHandle
                );
                appData.state = APP_STATE_BMX055_SET_MAG_RESET2;
            }
             
            break;
            
        case APP_STATE_BMX055_SET_MAG_RESET2:
 
 //           if (appData.transferStatus == APP_TRANSFER_STATUS_SUCCESS )
//            {
            
            // An ack signal doesn't appear after 0x83 in case of soft reset
            // As a result, appData.transferStatus is set to be APP_TRANSFER_STATUS_ERROR
            // wait 5 msec here, so that the soft reset signal (without last ack singnal) is fully sent  
            DelayMs(5);
                
                appData.transferStatus = APP_TRANSFER_STATUS_IN_PROGRESS; 
            
                TxData[0]= 0x4b;  // Select Mag register
                TxData[1]= 0x01;  // Soft reset
            
                DRV_I2C_WriteTransferAdd(appData.drvI2CHandle, MAG_ADDR,
                    (void *)&TxData, 2,
                    &appData.transferHandle
                );
                appData.state = APP_STATE_BMX055_SET_MAG_ODR;
 //           }
             
            break;

        case APP_STATE_BMX055_SET_MAG_ODR:
            
            if (appData.transferStatus == APP_TRANSFER_STATUS_SUCCESS)
            {            
                appData.transferStatus = APP_TRANSFER_STATUS_IN_PROGRESS; 
            
                TxData[0] =0x4c;  // Select Mag register
                TxData[1]= 0x00;  // Normal Mode, ODR = 10 Hz
            
                DRV_I2C_WriteTransferAdd(appData.drvI2CHandle, MAG_ADDR,
                    (void *)&TxData, 2,
                    &appData.transferHandle
                );

                appData.state = APP_STATE_BMX055_SET_MAG_ENABLE;
            }
            
            break;

        case APP_STATE_BMX055_SET_MAG_ENABLE:
 
            if (appData.transferStatus == APP_TRANSFER_STATUS_SUCCESS)
            {
                appData.transferStatus = APP_TRANSFER_STATUS_IN_PROGRESS; 
            
                TxData[0]= 0x4e;  // Select Mag register
                TxData[1]= 0x84;  // X, Y, Z-Axis enabled
            
                DRV_I2C_WriteTransferAdd(appData.drvI2CHandle, MAG_ADDR,
                    (void *)&TxData, 2,
                    &appData.transferHandle
                );
                appData.state = APP_STATE_BMX055_SET_MAG_REP_XY;
            }
             
            break;
            
        case APP_STATE_BMX055_SET_MAG_REP_XY:
 
            if (appData.transferStatus == APP_TRANSFER_STATUS_SUCCESS)
            {
                appData.transferStatus = APP_TRANSFER_STATUS_IN_PROGRESS; 
            
                TxData[0]= 0x51;  // Select Mag register
                TxData[1]= 0x04;  // No. of Repetitions for X-Y Axis = 9
            
                DRV_I2C_WriteTransferAdd(appData.drvI2CHandle, MAG_ADDR,
                    (void *)&TxData, 2,
                 &appData.transferHandle
                );
                appData.state = APP_STATE_BMX055_SET_MAG_REP_Z;
            }
             
            break;     
            
        case APP_STATE_BMX055_SET_MAG_REP_Z:
 
            if (appData.transferStatus == APP_TRANSFER_STATUS_SUCCESS)
            {
                appData.transferStatus = APP_TRANSFER_STATUS_IN_PROGRESS; 
            
                TxData[0]= 0x52;  // Select Mag register
                TxData[1]= 0x16;  // No. of Repetitions for Z-Axis = 15
            
                DRV_I2C_WriteTransferAdd(appData.drvI2CHandle, MAG_ADDR,
                    (void *)&TxData, 2,
                 &appData.transferHandle
                );
                appData.state = APP_STATE_BMX055_SET_END;
            }
             
            break;                        
            
        case APP_STATE_BMX055_SET_END:
 
            if (appData.transferStatus == APP_TRANSFER_STATUS_SUCCESS)
            {
                appData.state = APP_STATE_SCHEDULE_READ;
            }
             
            break;

///////////////////////////////////////////////////////            
        
        case APP_STATE_SCHEDULE_READ:

            if(APP_StateReset())
            {
                break;
            }

            /* If a read is complete, then schedule a read
             * else wait for the current read to complete */

            appData.state = APP_STATE_WAIT_FOR_READ_COMPLETE;
            if(appData.isReadComplete == true)
            {
                appData.isReadComplete = false;
                appData.readTransferHandle =  USB_DEVICE_CDC_TRANSFER_HANDLE_INVALID;

                USB_DEVICE_CDC_Read (USB_DEVICE_CDC_INDEX_0,
                        &appData.readTransferHandle, appData.cdcReadBuffer,
                        APP_READ_BUFFER_SIZE);
                
                if(appData.readTransferHandle == USB_DEVICE_CDC_TRANSFER_HANDLE_INVALID)
                {
                    appData.state = APP_STATE_ERROR;
                    break;
                }
            }

            break;

        case APP_STATE_WAIT_FOR_READ_COMPLETE:

            if(APP_StateReset())
            {
                break;
            }

            /* Check if a character was received or a switch was pressed.
             * The isReadComplete flag gets updated in the CDC event handler. */

            if(appData.isReadComplete)
            {
                appData.state = APP_STATE_BMX055_ACCL_READ;
            }

            break;


       case APP_STATE_BMX055_ACCL_READ:
        {  

            appData.transferStatus = APP_TRANSFER_STATUS_IN_PROGRESS;

            TxData[0]=0x02;//address of register
   
            /* Add a request to read data from BMX055. */
            DRV_I2C_WriteReadTransferAdd(appData.drvI2CHandle, ACCL_ADDR,
                (void *)&TxData[0], 1, (void *)&RxData[0], 6,
                &appData.transferHandle
            );
            
                appData.state = APP_STATE_BMX055_GYRO_READ;
             
             break;
         }
    
         case APP_STATE_BMX055_GYRO_READ:
         {
            if (appData.transferStatus == APP_TRANSFER_STATUS_SUCCESS)
            {
                xAccl = ((uint16_t)RxData[1] << 4) | (RxData[0] >> 4);
                if (xAccl > 2047)  xAccl -= 4096;
                
                yAccl = ((uint16_t)RxData[3] << 4) | (RxData[2] >> 4);
                if (yAccl > 2047)  yAccl -= 4096;
        
                zAccl = ((uint16_t)RxData[5] << 4) | (RxData[4] >> 4);
                if (zAccl > 2047)  zAccl -= 4096;
        
                xAccl = xAccl * 0.00957; // range +-2g (2/2^11 * 9.8)
                yAccl = yAccl * 0.00957; // range +-2g
                zAccl = zAccl * 0.00957; // range +-2g

                appData.transferStatus = APP_TRANSFER_STATUS_IN_PROGRESS;

                TxData[0]=0x02;//address of register
   
                /* Add a request to read data from BMX055. */
                DRV_I2C_WriteReadTransferAdd(appData.drvI2CHandle, GYRO_ADDR,
                        (void *)&TxData[0], 1, (void *)&RxData[0], 6,
                        &appData.transferHandle
                );
                        
                appData.state = APP_STATE_BMX055_MAG_READ;

            }
            break;
         }            

         case APP_STATE_BMX055_MAG_READ:
         {
            if (appData.transferStatus == APP_TRANSFER_STATUS_SUCCESS)
            {
                  // Convert the data
                xGyro = ((uint16_t)RxData[1] << 8) | RxData[0];
                if (xGyro > 32767)  xGyro -= 65536;
      
                yGyro = ((uint16_t)RxData[3] << 8) | RxData[2];
                if (yGyro > 32767)  yGyro -= 65536;
                
                zGyro = ((uint16_t)RxData[5] << 8) | RxData[4];
                if (zGyro > 32767)  zGyro -= 65536;
        
                xGyro = xGyro * 0.00381; //  Full scale = +/- 125 degree/s (125/2^15)
                yGyro = yGyro * 0.00381; //  Full scale = +/- 125 degree/s (125/2^15)
                zGyro = zGyro * 0.00381; //  Full scale = +/- 125 degree/s (125/2^15) 

                 appData.transferStatus = APP_TRANSFER_STATUS_IN_PROGRESS;

                TxData[0]=0x42;//address of register
   
                /* Add a request to read data from BMX055. */
                DRV_I2C_WriteReadTransferAdd(appData.drvI2CHandle, MAG_ADDR,
                        (void *)&TxData[0], 1, (void *)&RxData[0], 6,
                        &appData.transferHandle
                );
                        
                appData.state = APP_STATE_BMX055_READ_END;

            }
            break;
         }  

         case APP_STATE_BMX055_READ_END:
         {
            if (appData.transferStatus == APP_TRANSFER_STATUS_SUCCESS)
            {
                // Convert the data
                xMag = ((int16_t)RxData[1] <<5) | (RxData[0]>>3);
                if (xMag > 4095)  xMag -= 8192;
                 
                yMag = ((int16_t)RxData[3] <<5) | (RxData[2]>>3);
                if (yMag > 4095)  yMag -= 8192;
                
                zMag = ((int16_t)RxData[5] <<7) | (RxData[4]>>1);
                if (zMag > 16383)  zMag -= 32768;
              
                xMag = xMag * 0.317; //  Full scale = +/- 1300 uT (1300/2^12)
                yMag = yMag * 0.317; //  Full scale = +/- 1300 uT (1300/2^12)
                zMag = zMag * 0.153; //  Full scale = +/- 2500 uT (2500/2^14) 
                
                appData.state = APP_STATE_SCHEDULE_WRITE;

            }
            break;
         }  
  
         
        case APP_STATE_SCHEDULE_WRITE:

            if(APP_StateReset())
            {
                break;
            }

            /* Setup the write */

            appData.writeTransferHandle = USB_DEVICE_CDC_TRANSFER_HANDLE_INVALID;
            appData.isWriteComplete = false;
            appData.state = APP_STATE_WAIT_FOR_WRITE_COMPLETE;

            char a[80]={0};
            
            if(appData.cdcReadBuffer[0] != 'a' && appData.cdcReadBuffer[0] != 'g' && appData.cdcReadBuffer[0] != 'm')
            {
                appData.state = APP_STATE_SCHEDULE_READ;
                break;
            }
                
            if(appData.cdcReadBuffer[0] == 'a')
            sprintf(a,"xA=%5.2f  y=%5.2f  z=%5.2f \r\n", xAccl, yAccl, zAccl);
            else if(appData.cdcReadBuffer[0] == 'g')
            sprintf(a,"xG=%5.2f  y=%5.2f  z=%5.2f \r\n", xGyro, yGyro, zGyro);
            else
            sprintf(a,"xM=%5.2f  y=%5.2f  z=%5.2f \r\n", xMag, yMag, zMag);
            
            memcpy(appData.cdcWriteBuffer, a,strlen(a));
            
            USB_DEVICE_CDC_Write(USB_DEVICE_CDC_INDEX_0,
                    &appData.writeTransferHandle,
                    appData.cdcWriteBuffer, strlen(a),
                    USB_DEVICE_CDC_TRANSFER_FLAGS_DATA_COMPLETE);
                        
            break;

        case APP_STATE_WAIT_FOR_WRITE_COMPLETE:

            if(APP_StateReset())
            {
                break;
            }

            /* Check if a character was sent. The isWriteComplete
             * flag gets updated in the CDC event handler */

            if(appData.isWriteComplete == true)
            {
                appData.state = APP_STATE_SCHEDULE_READ;
            }

            break;

        case APP_STATE_ERROR:
        default:
            
            break;
    }
}

/*******************************************************************************
 End of File
 */

