/*******************************************************************************
  MPLAB Harmony Application Header File

  Company:
    Microchip Technology Inc.

  File Name:
    app.h

  Summary:
    This header file provides prototypes and definitions for the application.

  Description:
    This header file provides function prototypes and data type definitions for
    the application.  Some of these are required by the system (such as the
    "APP_Initialize" and "APP_Tasks" prototypes) and some of them are only used
    internally by the application (such as the "APP_STATES" definition).  Both
    are defined here for convenience.
*******************************************************************************/

#ifndef _APP_H
#define _APP_H

// *****************************************************************************
// *****************************************************************************
// Section: Included Files
// *****************************************************************************
// *****************************************************************************

#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdlib.h>
#include "configuration.h"
#include "definitions.h"

// DOM-IGNORE-BEGIN
#ifdef __cplusplus  // Provide C++ Compatibility

extern "C" {

#endif
// DOM-IGNORE-END

// *****************************************************************************
// *****************************************************************************
// Section: Type Definitions
// *****************************************************************************
// ***************************************************************************** 
#define APP_READ_BUFFER_SIZE                                512

// *****************************************************************************
/* Application States

  Summary:
    Application states enumeration

  Description:
    This enumeration defines the valid application states.  These states
    determine the behavior of the application at various times.
*/
typedef enum
{
    APP_TRANSFER_STATUS_IN_PROGRESS,
    APP_TRANSFER_STATUS_SUCCESS,
    APP_TRANSFER_STATUS_ERROR,
    APP_TRANSFER_STATUS_IDLE,

} APP_TRANSFER_STATUS;

typedef enum
{
    /* Application's state machine's initial state. */
    APP_STATE_INIT=0,

    /* Application waits for device configuration*/
    APP_STATE_WAIT_FOR_CONFIGURATION,

    /* Initialization for I2C */            
    APP_STATE_I2C_INIT,
            
    /* The application sets a ACCL RANGE of BMX055 */
    APP_STATE_BMX055_SET_ACCL_RANGE,
            
    /* The application sets a ACCL BW of BMX055 */
    APP_STATE_BMX055_SET_ACCL_BW, 
            
    /* The application sets a ACCL LPW of BMX055 */
    APP_STATE_BMX055_SET_ACCL_LPW,
            
    /* The application sets a GYRO RANGE of BMX055 */
    APP_STATE_BMX055_SET_GYRO_RANGE,
            
    /* The application sets a GYRO BW of BMX055 */
    APP_STATE_BMX055_SET_GYRO_BW, 
            
    /* The application sets a GYRO LPM1 of BMX055 */
    APP_STATE_BMX055_SET_GYRO_LPM1,
            
    /* The application sets a MAG soft RESET of BMX055 */
    APP_STATE_BMX055_SET_MAG_RESET1,            
            
    /* The application sets a MAG Soft RESET of BMX055 */
    APP_STATE_BMX055_SET_MAG_RESET2,            

    /* The application sets a MAG ODR of BMX055 */
    APP_STATE_BMX055_SET_MAG_ODR,

    /* The application sets a MAG ODR of BMX055 */
    APP_STATE_BMX055_SET_MAG_ENABLE,
            
    /* The application sets a MAG Repetitions for X-Y of BMX055 */
    APP_STATE_BMX055_SET_MAG_REP_XY,

    /* The application sets a MAG Repetitions for Z of BMX055 */
    APP_STATE_BMX055_SET_MAG_REP_Z,
            
            
    APP_STATE_BMX055_SET_END,

    /* Wait for a character receive */
    APP_STATE_SCHEDULE_READ,

    /* A character is received from host */
    APP_STATE_WAIT_FOR_READ_COMPLETE,

            
APP_STATE_BMX055_ACCL_READ,
//APP_STATE_BMX055_ACCL_READ_END,            
APP_STATE_BMX055_GYRO_READ,
APP_STATE_BMX055_MAG_READ,
APP_STATE_BMX055_READ_END,

  APP_STATE_BMX055_READ_MAG_RESET,          
    /* Wait for the TX to get completed */
    APP_STATE_SCHEDULE_WRITE,

    /* Wait for the write to complete */
    APP_STATE_WAIT_FOR_WRITE_COMPLETE,

    /* Application Error state*/
    APP_STATE_ERROR

} APP_STATES;


// *****************************************************************************
/* Application Data

  Summary:
    Holds application data

  Description:
    This structure holds the application's data.

  Remarks:
    Application strings and buffers are be defined outside this structure.
 */

typedef struct
{
    /* Device layer handle returned by device layer open function */
    USB_DEVICE_HANDLE deviceHandle;

    /* Application's current state*/
    APP_STATES state;

    /* Set Line Coding Data */
    USB_CDC_LINE_CODING setLineCodingData;

    /* Device configured state */
    bool isConfigured;

    /* Get Line Coding Data */
    USB_CDC_LINE_CODING getLineCodingData;

    /* Control Line State */
    USB_CDC_CONTROL_LINE_STATE controlLineStateData;

    /* Read transfer handle */
    USB_DEVICE_CDC_TRANSFER_HANDLE readTransferHandle;

    /* Write transfer handle */
    USB_DEVICE_CDC_TRANSFER_HANDLE writeTransferHandle;

    /* True if a character was read */
    bool isReadComplete;

    /* True if a character was written*/
    bool isWriteComplete;

    /* Flag determines SOF event occurrence */
    bool sofEventHasOccurred;

    /* Break data */
    uint16_t breakData;

    /* Application CDC read buffer */
    uint8_t * cdcReadBuffer;

    /* Application CDC Write buffer */
    uint8_t * cdcWriteBuffer;

    /* Number of bytes read from Host */ 
    uint32_t numBytesRead; 
    
    /* I2C Driver Handle */
    DRV_HANDLE drvI2CHandle;

    /* Ready transfer handle */
    DRV_I2C_TRANSFER_HANDLE transferHandle;

    /* Transfer status */
    volatile APP_TRANSFER_STATUS transferStatus;
    
} APP_DATA;

// *****************************************************************************
// *****************************************************************************
// Section: Application Callback Routines
// *****************************************************************************
// *****************************************************************************
/* These routines are called by drivers when certain events occur.
*/

// *****************************************************************************
// *****************************************************************************
// Section: Application Initialization and State Machine Functions
// *****************************************************************************
// *****************************************************************************

/*******************************************************************************
  Function:
    void APP_Initialize ( void )

  Summary:
     MPLAB Harmony application initialization routine.

  Description:
    This function initializes the Harmony application.  It places the 
    application in its initial state and prepares it to run so that its 
    APP_Tasks function can be called.

  Precondition:
    All other system initialization routines should be called before calling
    this routine (in "SYS_Initialize").

  Parameters:
    None.

  Returns:
    None.

  Example:
    <code>
    APP_Initialize();
    </code>

  Remarks:
    This routine must be called from the SYS_Initialize function.
*/

void APP_Initialize ( void );


/*******************************************************************************
  Function:
    void APP_Tasks ( void )

  Summary:
    MPLAB Harmony Demo application tasks function

  Description:
    This routine is the Harmony Demo application's tasks function.  It
    defines the application's state machine and core logic.

  Precondition:
    The system and application initialization ("SYS_Initialize") should be
    called before calling this.

  Parameters:
    None.

  Returns:
    None.

  Example:
    <code>
    APP_Tasks();
    </code>

  Remarks:
    This routine must be called from SYS_Tasks() routine.
 */

void APP_Tasks( void );

//DOM-IGNORE-BEGIN
#ifdef __cplusplus
}
#endif
//DOM-IGNORE-END

#endif /* _APP_H */

/*******************************************************************************
 End of File
 */

